searchData={"items":[{"type":"module","title":"Mar","doc":"Using `Mar` makes the module a route.\nIt injects `Mar.Route` protocol and a struct.\nSet a `path` or the default is \"/\".\nAdd parameters in the path or in the `params` list.\n\n```elixir\ndefmodule MyApp do\n  use Mar, path: \"/post/:id\", params: [:likes, :comment]\nend\n```\n\nName your function after the HTTP method.\nTake HTTP parameters from a map.\nReturn a response body with text.\n\n```elixir\ndef get(%{id: id}) do\n  \"You are reading #{id}\"\nend\n```\n\nReturning a map will send a response in JSON.\n\n```elixir\ndef post(%{id: id, comment: comment}) do\n  %{\n    id: id,\n    comment: comment\n  }\nend\n```\n\nReturn a tuple to set HTTP status and headers.\n\n```elixir\ndef delete(%{id: _id}) do\n  # {status, header, body}\n  {301, [location: \"/\"], nil}\nend\n```\n\n`Mar.Route` protocol lets you access `Plug.Conn`.\n\n```elixir\ndefimpl Mar.Route do\n  # Mar.Route.MyApp\n\n  def before_action(route) do\n    # Access `route.conn` before the actions you have defined.\n    route\n  end\n\n  def after_action(route) do\n    # Access `route.conn` after the actions you have defined.\n    route\n  end\nend\n```","ref":"Mar.html"},{"type":"protocol","title":"Mar.Route","doc":"`Mar.Route` protocol lets you access `Plug.Conn`.\n\n```elixir\ndefmodule MyApp do\n  # ...\n\n  defimpl Mar.Route do\n    # Mar.Route.MyApp\n\n    def before_action(route) do\n      # Access `route.conn` before the actions you have defined.\n      route\n    end\n\n    def after_action(route) do\n      # Access `route.conn` after the actions you have defined.\n      route\n    end\n  end\nend\n\n```","ref":"Mar.Route.html"},{"type":"function","title":"Mar.Route.after_action/1","doc":"","ref":"Mar.Route.html#after_action/1"},{"type":"function","title":"Mar.Route.before_action/1","doc":"","ref":"Mar.Route.html#before_action/1"},{"type":"type","title":"Mar.Route.t/0","doc":"All the types that implement this protocol.","ref":"Mar.Route.html#t:t/0"},{"type":"extras","title":"README","doc":"![mar](priv/mar.png \"Mar\")\n\n# Mar — Simple Web in Elixir\n\n[Video Demo](https://youtu.be/lcec6JbynYo)\n\nMar is a simple web framework written in Elixir functional programming language. The design of the library aims to provide a Flask-like experience for Elixir developers. It is a layer of abstraction on top of [Plug](https://github.com/elixir-plug/plug) powered by [Bandit](https://github.com/mtrudel/bandit).","ref":"readme.html"},{"type":"extras","title":"Design - README","doc":"```mermaid\ngraph\n\nSupervisor\nBandit\n\nsubgraph legend\n  direction TB\n  modules\n  protocols[[protocols]]\n  impls(impls)\nend\n\nsubgraph mar\n  Mar\n  Mar.Route[[Mar.Route]]\n  Mar.Router\nend\n\nsubgraph my_app\n  MyApp\n  Mar.Route.MyApp(Mar.Route.MyApp)\nend\n\nMar.Router -- \"__protocol__(:impls)\" --> Mar.Route\nMar.Router -- \"%MyApp{}\" --> Mar.Route.MyApp\nMar.Router -- apply --> MyApp\nMyApp -- use --> Mar\nMyApp -- start_link --> Supervisor\nMyApp -- defimpl --> Mar.Route.MyApp\nSupervisor -- child_spec --> Mar\nSupervisor -- \":start\" --> Bandit\nBandit -- \":plug\" --> Mar.Router\n\n```\n\nThe design goal for Mar is to make it _transparent_ to the user so someone who just covered the basics of the language can jump right into building a web application. Elixir offers a powerful and aesthetic syntax built-in, such as multi-clause function definitions with parameters pattern matching module with struct definition, `@behaviour` and `Protocol` for polymorphism, and `use` macro that injects code into the module. Mar attempts to fully utilise these native features to provide a simple and intuitive API for building web applications.\n\nFor the user, such brevity means typing in `use Mar` makes all the connections and sets up reasonable defaults so a web application is defined without an extra file in the `lib/`. With object-oriented programming, it's as simple as passing an object. That's what we see from Flask, Sinatra or Express. The object can encapsulate the state and operation of the application without too much ceremony on the user's codespace. But for the functional programming paradigm, it usually spills out to the user modules. To manage this, Mar has to keep as much as possible in the library. And it takes accessing the module information of the user code from the library.\n\nI was surprised by the level of frustration such an approach can cause. At first, I thought it was me that I couldn't come up with an elegant solution right up. But as I dug deeper, I found the trail of lamentation. Most libraries resort to taking some configuration for a module, so the user can manually report to the library. Other than that, the ambition of the community has reached scanning all the modules in the user project and saving them in an [ETS](https://www.erlang.org/doc/man/ets.html) or an [Agent](https://hexdocs.pm/elixir/Agent.html). I tried it and it was a convoluted, error-prone, and hard-to-maintain solution, at least for my technical capability.\n\n[Protocol](https://hexdocs.pm/elixir/Protocol.html) is an often underappreciated feature of the language. And it sheds light on the problem. Protocols do [module consolidation](https://hexdocs.pm/elixir/Protocol.html#module-consolidation) in order to enhance the performance. And the protocol can reach the consolidated modules with `__protocol__(:impls)`. If Mar can slip in protocol implementation code into the user module with `use Mar`, it will automatically report the module to the protocol. And Mar can access the module without extra configuration. This comes with a nice side effect. Since `defimpl` takes structs anyway, saving information in the struct is a natural way to keep the module information. And the protocol can work as a liminal codespace instead of being a tool for module registration.\n\nSo that's exactly the way Mar is implemented. Each user module is a unique implementation of `Mar.Route`. It provides some default `defimpl` functions and `Mar.__using__/1` injects them alongside `defstruct` handling options. This way, `use Mar` makes these things possible:\n\n1. Register the module to the `Mar.Route`.\n2. Save information in the `%MyApp{}` struct.\n3. Interact with the library through `Mar.Route.MyApp`\n\n`use Mar` takes `path` and `params` options. `path` is the URL path of the route where path parameters are annotated with `:`. They are parsed and appended to the existing `params`.\n\n`Mar` also exports `Mar.child_spec/1` so the user can start the supervision tree just by adding `Mar` as a child. The child spec starts Bandit with `Mar.Plug` as a general pipeline which goes down to `Mar.Router` where path-matching, parameter parsing, and action dispatching are done.\n\nPath-matching algorithm is far from optimised. It's a few iterations over the list of routes. [Stream](https://hexdocs.pm/elixir/Stream.html) is applied where applicable to alleviate the performance. Once the path finds the module of responsibility, path parameters are parsed to the [Plug.Conn](https://hexdocs.pm/plug/Plug.Conn.html) first, where they merge over query parameters and body parameters. Then parameters are selectively loaded to the struct. The HTTP method is loaded as well for the action name. Then before the action is dispatched, `conn` is loaded to the struct. This allows `Mar.Route` to expose the entire connection to the user module. After the action is dispatched, it exposes the connection once more. And finally, the response gets sent.","ref":"readme.html#design"},{"type":"extras","title":"Installation - README","doc":"Create a project with supervision tree:\n\n```sh\n$ mix new my_app --sup\n$ cd my_app\n```\n\nAdd Mar to your dependencies in:\n\n```elixir\n# mix.exs\ndefp deps do\n  [\n    {:mar, \"~> 0.2.0\"}\n  ]\nend\n```\n\nAdd Mar as a child:\n\n```elixir\n# lib/my_app/application.ex\ndef start(_type, _args) do\n  children = [\n    Mar\n  ]\n  # ...\nend\n```\n\nAdd Mar to your module to make it a route:\n\n```elixir\n# lib/my_app/route.ex\ndefmodule MyApp.Page do\n  use Mar\n\n  def get() do\n    \"Hello, world!\"\n  end\nend\n```\n\nSpin up a server:\n\n```sh\n$ mix run --no-halt\n```\n\n```sh\n$ curl localhost:4000\n# Hello, world!\n```","ref":"readme.html#installation"},{"type":"extras","title":"Use - README","doc":"Using `Mar` makes the module a route.\nIt injects `Mar.Route` protocol and a struct.\nSet a `path` or the default is \"/\".\nAdd parameters in the path or in the `params` list.\n\n```elixir\ndefmodule MyApp do\n  use Mar, path: \"/post/:id\", params: [:likes, :comment]\nend\n```\n\nName your function after the HTTP method.\nTake HTTP parameters from a map.\nReturn a response body with text.\n\n```elixir\ndef get(%{id: id}) do\n  \"You are reading #{id}\"\nend\n```\n\nReturning a map will send a response in JSON.\n\n```elixir\ndef post(%{id: id, comment: comment}) do\n  %{\n    id: id,\n    comment: comment\n  }\nend\n```\n\nReturn a tuple to set HTTP status and headers.\n\n```elixir\ndef delete(%{id: _id}) do\n  # {status, header, body}\n  {301, [location: \"/\"], nil}\nend\n```\n\n`Mar.Route` protocol lets you access `Plug.Conn`.\n\n```elixir\ndefimpl Mar.Route do\n  # Mar.Route.MyApp\n\n  def before_action(route) do\n    # Access `route.conn` before the actions you have defined.\n    route\n  end\n\n  def after_action(route) do\n    # Access `route.conn` after the actions you have defined.\n    route\n  end\nend\n```","ref":"readme.html#use"},{"type":"extras","title":"Reference - README","doc":"- [Mar on HexDocs](https://hexdocs.pm/mar)\n- [Proposal on Elixir Forum](https://elixirforum.com/t/62885)","ref":"readme.html#reference"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}